#### Y: binary disease outcome
#### Z: binary treatment status
#### S: biomarker(s) 
#### start: initial value for risk model parameter
#### fpr: points on the ROC curve
#### threshold: thresholds for Q(s) = Risk0(s) - Risk1(s), TPF, FPF, and CDF_Q curves
#### parsen: parameter for sensitivity analyis  

# Models(eqn 3) g{P(C <= k | S)} = [1,q(S)] %*% a_k k=1,2,3, a_k = (a_k0,..,)' , dim(a_k) = 1 + dim(q(S)) 
# (a_1,a_2) are the only indep. parameters since 1 = sum(P{C = k | S}) for k = 1,2,3.

# Model(eqn 4) g[P{Y(Z) = 1 | S} = th_0 + th_1*Z + th_2'q(S) + th_3'q(S)*Z
# theta = (th_0,th_1,th_2,th_3)

# Computation
# theta = theta[t1,t2] = [(th_0,th_2'),(th_1,th_3')] (note the order) corresponds to model matrix [1,S,Z,Z*S] = [(1,S),Z*(1,S)]
# alpha = alpha[a1,a2] = [(a_10,a1_'),(a20,a2_')] corresponds to matrix [(1,S),(1,S)]


# Below and in the R code 'S' contains the intercept: 'S' = [1,q(S)].
# risk(theta), compute risk via risk-scale parameter 'theta': theta = [(th0,th2')' , (th1,th3')']
# S = [1,q(S)]                         Z = 0           Z = 1     
# risk1 = P{Y(Z) = 1 | S} = linkinv(S%*%theta[,1] + S%*%theta[,2])
# risk0 = P{Y(Z) = 0 | S} = linkinv(S%*%theta[,1])

# risk(alpha): compute risk via parameter 'alpha': a = [a_1,a_2] is dim(S) x 2	
# S = [1,q(S)]
# risk0 = P(C <= 2 | S) = linkinv(S %*% a[,2])
# risk1 = P(C <= 1 | S) = linkinv(S %*% a[,1])

# 'treatment' is a right-hand formula of the form '~Z'. It is needed because  
# a formula such as Y~Z*(S1+..+Sd) produces [(Intercept), Z, S1,..,Sd,    Z*S1,..,Z*Sd] 
# a formula such as Y~(S1+..+Sd)*Z produces [(Intercept),    S1,..,Sd, Z, Z*S1,..,Z*Sd]
# but we want to decompose a model matrix as [(1,S) , Z*(1,S)]



try.warn <- function(expr, silent=TRUE){
  res <- try(expr, silent)
  if(inherits(res,'try-error')) warning(attr(res,'condition')$message)
  invisible(res)
}

logit <-  function(x)log(x/(1.0-x))
expit <- function (x) {1.0/(1.0+exp(-x))}


# in paper th0 th1 th2 th3
# in code: (th0 th1)
#          (th2 th3)

# in paper 

# g[P{Y(Z)=1|S}] = th0 + th1*Z + th2'q(S) + th3'q(S) * Z
# input:  theta = (th0,th2',th1,th3')
# output: alpha = rbind(alpha_1',alpha_2')   
alpha.theta <- function(p){
# max number of categories in model (eqn 2) is C \in {1,2,3} 
# but only indep. parameters are for for C <= 1 and C <= 2 
    maxC <- 2
    d<- length(p)
    if((d %% 2) || (d < 4)) stop("incorrect dimension of theta: ",d,"\ndim(theta) should be even, >3 and corresponds to model g[P{Y(Z)=1|S}] = th0 + th1*Z + th2'q(S) + th3'q(S) * Z")	
    dS <- (d - 2) / maxC	
    theta <- matrix(ncol = maxC,data = p)
    a     <- theta * NA   
    NM <- attr(p,'NAMES')
    if(length(NM)){
        attr(a,'NAMES') <- NM
        colnames(a) <- c("C=1","C=2")
        rownames(a) <- NM[-1]
    }
    a[1,1]  <- theta[1,1] + theta[1,2]   #alpha_1 = th0 + th1
    a[-1,1] <- theta[-1,1] + theta[-1,2] #beta_1 = th2 + th3
    a[1,2]  <- theta[1,1]                #alpha_2 = th0
    a[-1,2] <- theta[-1,1]               #beta_2 = th2
    a
}

# alpha below (denoted by 'a') does not include alpha_3 (k=3)
# input:  alpha = c(cbind(alpha_1,alpha_2)) 
# output: theta = (th0,th2',th1,th3')            
theta.alpha <- function(p){
# max number of categories in model (eqn 2) is C \in {1,2,3} 
# but only parameters that are independent are for C <= 1 and C <= 2 
    maxC <- 2
    d<- length(p)
    if((d %% 2) || (d < 4)) stop("incorrect dimension of alpha: ",d,"\ndim(alpha) should be even, >3 and corresponds to model g[P{C<=k|S}] = a0 + a1*q(S)")	
    dS <- (d - 2) / maxC
    a <- matrix(ncol = maxC,data = p)
    theta <- a * NA
    NM <- attr(p,'NAMES')
    if(length(NM)){
        attr(theta,'NAMES') <- NM
        colnames(theta) <- c("",paste(NM[1],":",sep = ""))
        rownames(theta) <- NM[-1]
    }
    theta[1,1]  <- a[1,2]
    theta[-1,1] <- a[-1,2]
    theta[1,2]  <- a[1,1]  - a[1,2]
    theta[-1,2] <- a[-1,1] - a[-1,2]	
    theta
}

# Risk Model
# paper  : g(P{Y(Z) = 1 | S}) = t0 + t1*Z + t2*S + t3*Z*S
# code   : g(P{Y(Z) = 1 | S}) = t1 + t2*S + t3*Z + t4*Z*S
# theta  : (theta1,theta3) matrix of dimension (1 + q , 2)
#          (theta2,theta4)
# input: x = [1,S]_{n x (1+q)} , theta = (1,S,Z,Z*S)_{2(1+q)}, q = dim(S),no. of markers
# output [risk0,risk1] = linkinv([1,S]%*%theta[1:(1+q)]),linkinv([1,S]%*%theta[-(1:(1+q))]),
risk.theta<-function(x,theta,linkinv = expit)
{
    if(length(dim(theta)) != 2) theta <- matrix(ncol = 2,data = theta)	
    eta.theta <- x %*% theta
    eta.theta[,2] <- eta.theta[,1] + eta.theta[,2]
    apply(eta.theta,2,linkinv)
}

# Cumulative link model for potential outcomes
# paper  : g(P{C <= k | S}) = ak + bk * S, dim(ak) = 1, dim(bk) = q, q = dim(S),no. of markers
# code   : (alpha1,alpha2,alpha3,alpha4) = (a1,b1,a2,b2) 

# input:
# x = [1,S]_{n x (1+q)}
# alpha  : (alpha1,alpha3) matrix of dimension (1 + q , 2)
#          (alpha2,alpha4)
# output [risk0,risk1]
risk.alpha<-function(x,alpha,linkinv = expit)
{
    if(length(dim(alpha)) != 2) alpha <- matrix(ncol = 2,data = alpha)
    apply(x %*% alpha[,c(2,1)],2,linkinv)
}

get.risk<-function(x,theta = NULL,alpha = NULL,linkinv = expit)
{
    if(missing(x))stop("markers 'x' must be given.") 	
    if(is.null(theta) && is.null(alpha)) stop("Either alpha or theta must be given.")
    if(is.character(linkinv)) linkinv <- get(linkinv, mode = "function", envir = parent.frame())

    NM <- NULL
    if(length(alpha)){
        NM <- attr(alpha,'NAMES')
    # risk(alpha), compute risk via parameters a_k, k=1,2
        if(length(dim(alpha)) != 2) alpha <- matrix(ncol = 2,data = alpha)
	    risk <- apply(x %*% alpha[,c(2,1)],2,linkinv)
    }else{
    # risk(theta), compute risk via risk-scale parameter 'theta' 
        NM <- attr(theta,'NAMES')
        if(length(dim(theta)) != 2) theta <- matrix(ncol = 2,data = theta)
		risk <- x %*% theta #[risk0 = (1,S),riskZ = (Z,Z*S)]
		risk[,2] <- risk[,1] + risk[,2]#[risk0 = (1,S),risk1 = (1,S) + (Z,Z*S)]
		risk <- apply(risk,2,linkinv)
    }
	if(!length(NM)) NM <- "tx" 
	colnames(risk) <- paste(NM[1],0:1,sep = "=")
    risk
}


delta.phi_gamma<-function(delta,S_gamma,risk1,phi_rho1,linkinv){mean(linkinv(delta + S_gamma) * risk1)-phi_rho1}
# expit vs built-in 'logit' linkinv gives different estimates at 'Inf' due to roundoff error 
# delta.phi_gamma<-function(delta,S_gamma,risk1,phi_rho1,linkinv){
    # y <- mean(linkinv(delta + S_gamma) * risk1)-phi_rho1
    # round(y,digits = 15)
# }


## Constrained likelihood
# risk_Z(S) = Risk_{(Z)}(S) = P{Y(Z)=1|S} = P{Y=1|Z,S} = Risk(S,Z) defines risk_0,risk_1 for z=0,1
# s = covariates without intercept - in code 'S' = [1,s]
# S_gamma = (delta,gamma)'(1,s)
# w = w(s,delta,gamma)
# determine delta by eqn 7: Integral{ w(s,d,g) * risk1(s) dF(s)} = phi * P{Y(z=1) = 1}
negloglik.tsm <- function(p,parsen = NULL,Y,Z,S,linkinv = expit,riskfun = risk.alpha,verbose = 0){
        
        if(verbose) cat(sprintf("%.10f ",p))    
        lik <- -Inf
		r <- riskfun(S,p,linkinv) 
        if(any(is.na(c(r)))){
            if(verbose)cat(" -lik ",-lik,"\n")
            return(-lik)
        }
        
        r0 <- r[,1]
        r1 <- r[,2]
        w<-0.0
        
        if(length(parsen)){
            phi <- parsen[1]
            if (phi > 0.0){
                gm<-as.matrix(parsen[-1]);# since S contains intercept but delta is handled separately
                S_gamma <- c(S[,-1,drop = FALSE] %*% gm)
                delta<-try(uniroot(delta.phi_gamma,lower=-100,upper=100,
                                        S_gamma = S_gamma,risk1 = r1,phi_rho1 = phi*mean(r1),
                                        linkinv = linkinv)$root,TRUE)            
                if (inherits(delta,'try-error')) {if(verbose)cat(" -lik ",-lik,"\n");return(Inf)}
                w <- linkinv(delta+S_gamma)
                if(any(!is.finite(w))){if(verbose)cat(" -lik ",-lik,"\n");return(-lik)}
            }        
        }
        
# constraints (6) in paper
      
# if w = 0.0 (monotonicity satisfied) then by eqn-3 in paper
# (i)   r0 = risk0 = linkinv(a2 + b2 %*% S) = Pr(C = 2 | S)        
# (ii)  r1 = risk1 = linkinv(a1 + b1 %*% S) = Pr(C = 1 | S)
# => constraints are: 
# 0 <= r0 - r1  : monotonicity constraint enforced
# 0 <= r0 <= 1
# 0 <= r1 <= 1 and 
# 1.0= Pr(C = 1 | S)+Pr(C = 2 | S)+Pr(C = 3 | S) => 0 <= 1-r0-r1 <= 1

# if w > 0 (monotonicity violated) then constraints are as below 

        if ((min(c(r,r0-(1.0-w)*r1,1.0-r0-w*r1)) < 0.0)) {
            if(verbose)cat(" -lik ",-lik,"\n");
            return(-lik)
        }

        lik<-sum(Y[Z==1]*log(r1[Z==1]))+sum((1.0 - Y[Z==1])*log(1.0 - r1[Z==1])) +
             sum(Y[Z==0]*log(r0[Z==0]))+sum((1.0 - Y[Z==0])*log(1.0 - r0[Z==0]))

#        if(verbose) cat(sprintf("%.10f",p)," -lik ",-lik,"\n");
        if(verbose)cat(" -lik ",-lik,"\n")
        return(-lik)
}  

## Unconstrained likelihood
loglik.tsm<-function(y,z,x,theta,linkinv){
		risk <- risk.theta(x,theta,linkinv)   
        if(any(is.na(c(risk))))return(-Inf)
        r0 <- risk[,1]
        r1 <- risk[,2]
        y <- as.numeric(y)
        z <- as.logical(z)
        lik<-sum(y[z]*log(r1[z]))+sum((1.0 - y[z])*log(1.0 - r1[z])) +
             sum(y[!z]*log(r0[!z]))+sum((1.0 - y[!z])*log(1.0 - r0[!z]))
        return(lik)
}    
    
    
    
## output if parsen != NULLL: auc,tpf.q,fpf.q,Rinvp.q,roc.q,TPF.Q,FPF.Q = FPF.Q,Q where 'q' = threshold, 'Q' = c(-Inf,sort(unique(Q))) 
## output if parsen  = NULLL: auc_u,auc_n,auc_h,roc_u,roc_n,roc_h,TPFn.Q,FPFn.Q,FPFh.Q,Q where 'q' = threshold, 'Q' = c(-Inf,sort(unique(Q))) 
## NOTE: it is important that if computation of a summary measure fails, the output for that measure must contain a vector of NAs of the expected length   
get.measures<-function(risk,S,parsen = NULL,fpr = seq(.1,.9,by = .1),threshold = seq(.1,.9,by = .1),linkinv = expit,na.rm = FALSE){

    Q <- c(risk[,1]-risk[,2])
    Q0 <- c(-Inf,sort(unique(Q)))  
    
    if(!length(parsen)){
        auc<-NA
        roc.q<-rep(NA,length(fpr))
        q  <- threshold	
        TPF.q<-FPF.q<-Rinvp.q<-rep(NA,length(q))		
        TPF.Q <- FPF.Q <- Q0*NA    
      
        mQ <- mean(Q)	
        if(all(Q >= 0) && mQ > 0){
            IQQ <- outer(Q,Q0,FUN = '>')
            QQ  <- outer(Q,rep(1.0,length(Q0)))
            TPF.Q <- colMeans(QQ * IQQ) / mQ
            FPF.Q <- colMeans((1.0 - QQ) * IQQ) / (1.0-mQ)
            roc   <- try(approx(FPF.Q,TPF.Q,fpr)$y,TRUE)
            if(!inherits(roc,'try-error'))roc.q <- roc
            
            auc<-(-1)*sum((TPF.Q[-1]+TPF.Q[-length(TPF.Q)])/2.0*diff(FPF.Q))          
            q <- threshold
            IQq <- outer(Q,q,FUN = '>')
            QQ  <- outer(Q,rep(1.0,length(q)))
            TPF.q <- colMeans(QQ * IQq) / mQ
            FPF.q <- colMeans((1.0 - QQ) * IQq) / (1.0-mQ)
            qq <- outer(rep(1.0,length(Q)),q)
            Rinvp.q<-colMeans(IQq <= qq)
        }
        measures <- list(auc=auc,tpf.q=TPF.q,fpf.q=FPF.q,Rinvp.q=Rinvp.q,roc=roc.q,TPF.Q = TPF.Q,FPF.Q = FPF.Q,Q = Q0)
        
        return (measures) 
    }

###########################################################################################################           
     
    w<-0
    phi <- parsen[1]
    delta <- NA
    if (phi > 0){
        gm<-parsen;gm[1] <- 0 # since S contains intercept but delta is handled separately
        S_gamma <- c(S %*% gm)
        phi_rho1 <- phi*mean(risk[,2])
        delta<-try(uniroot(delta.phi_gamma,lower=-100,upper=100,
                                S_gamma = S_gamma,risk1 = c(risk[,2]),phi_rho1 = phi_rho1,
                                linkinv = linkinv)$root)            
        if (inherits(delta,'try-error'))w<-NA else w<-linkinv(delta+S_gamma)
    }

   	rr1<-apply(cbind(-Q,risk[,2]*w),1,max)
	rr1<-apply(cbind(rr1,1-risk[,1]),1,min)
	rr2<-risk[,2]-rr1                      
	rr3<-risk[,1]-rr2
	rr4<-1-rr1-rr2-rr3
	index<-rr1==risk[,2]*w
	index<-ifelse(rr1==-Q,2,index)
	index<-ifelse(rr1==1-risk[,1],3,index)

    TPFn <- lapply(Q0,function(x){mean((rr3)*(Q>x),na.rm = na.rm)/mean(rr3,na.rm = na.rm)})
    TPFn <- unlist(TPFn)
    
    FPFu <- lapply(Q0,function(x){mean((rr2+rr4)*(Q>x),na.rm = na.rm)/mean(rr2+rr4,na.rm = na.rm)})
    FPFu <- unlist(FPFu)

    FPFh <- lapply(Q0,function(x){mean((rr1)*(Q>x),na.rm = na.rm)/mean(rr1,na.rm = na.rm)})
    FPFh <- unlist(FPFh)

    FPFn <- lapply(Q0,function(x){mean((rr1+rr2+rr4)*(Q>x),na.rm = na.rm)/mean(rr1+rr2+rr4,na.rm = na.rm)})    
    FPFn <- unlist(FPFn)
    
    AUCu <- sum((TPFn[-length(TPFn)]+TPFn[-1])/2*(-1)*diff(FPFu),na.rm = TRUE)
    AUCn <- sum((TPFn[-length(TPFn)]+TPFn[-1])/2*(-1)*diff(FPFn),na.rm = TRUE)
    AUCh <- sum((TPFn[-length(TPFn)]+TPFn[-1])/2*(-1)*diff(FPFh),na.rm = TRUE)

    ROCu <- try(approx(FPFu,TPFn,fpr)$y,TRUE);if(inherits(ROCu,'try-error')) ROCu <- rep(NA,length(fpr))
    ROCn <- try(approx(FPFn,TPFn,fpr)$y,TRUE);if(inherits(ROCn,'try-error')) ROCn <- rep(NA,length(fpr))    
    ROCh <- try(approx(FPFh,TPFn,fpr)$y,TRUE);if(inherits(ROCh,'try-error')) ROCh <- rep(NA,length(fpr)) 

    measures <- list(auc_u=AUCu,auc_n=AUCn,auc_h=AUCh,roc_u=ROCu,roc_n=ROCn,roc_h=ROCh,TPFn.Q=TPFn,FPFn.Q=FPFn,FPFh.Q=FPFh,Q = Q0)
###########################################################################################################                
    return(measures)                
}


# initial values ranked by likelihood
tsm.start <- function(Y,Z,S,parsen = NULL,link,linkinv,verbose = 0){
    Start           <- list()
    s               <- NCOL(S)
    theta0          <- rep(0.0,2 * s)

# zero vector    
    Start$zero      <- list(theta = theta0,loglik = -Inf,info = "zero vector")    
    # loglik <- try(loglik.tsm(Y,Z,S,Start$zero$theta,linkinv = linkinv),silent = TRUE)
    # if(!inherits(loglik,'try-error'))Start$zero$loglik <- loglik 
    loglik <- try(negloglik.tsm(Start$zero$theta,parsen = parsen,Y,Z,S,linkinv = linkinv,riskfun = risk.theta,verbose = verbose),silent = TRUE)
    if(!inherits(loglik,'try-error'))Start$zero$loglik <- -loglik 
    
# main-effects only via glm(), no interaction: ~1 + S + Z + 0*S*Z   
    theta <- theta0     
    Start$main      <- list(theta = theta0*NA,loglik = -Inf,info = "main effects: ~1 + S + Z + 0*S*Z")
    theta[1:(s+1)] <- try(glm.fit(cbind(S,Z),Y,family = binomial(link = link$name))$coefficients,silent = TRUE)
    if(!inherits(theta,'try-error')){
        Start$main$theta <- theta
        # loglik <- try(loglik.tsm(Y,Z,S,Start$main$theta,linkinv = linkinv),silent = TRUE)
        # if(!inherits(loglik,'try-error'))Start$main$loglik <- loglik 
        loglik <- try(negloglik.tsm(Start$main$theta,parsen = parsen,Y,Z,S,linkinv = linkinv,riskfun = risk.theta,verbose = verbose),silent = TRUE)
        if(!inherits(loglik,'try-error'))Start$main$loglik <- -loglik 
    }

# interaction via glm(): ~1 + S + Z + S*Z            
    Start$int       <- list(theta = theta0*NA,loglik = -Inf,info = "main effects with interaction: ~1 + S + Z + S*Z")    
    Start$int$theta <- try(glm.fit(cbind(S,Z*S),Y,family = binomial(link = link$name))$coefficients,silent = TRUE)
    if(!inherits(Start$int$theta,'try-error')){
        # loglik <- try(loglik.tsm(Y,Z,S,Start$int$theta,linkinv = linkinv),silent = TRUE)
        # if(!inherits(loglik,'try-error'))Start$int$loglik <- loglik 
        loglik <- try(negloglik.tsm(Start$int$theta,parsen = parsen,Y,Z,S,linkinv = linkinv,riskfun = risk.theta,verbose = verbose),silent = TRUE)
        if(!inherits(loglik,'try-error'))Start$int$loglik <- -loglik 
    }
    
# stratification via glm(): (~1 + S | Z = 0) , (~1 + S | Z = 1)         
    Start$strat   <- list(theta = theta0*NA,loglik = -Inf,info = "stratified: (~1 + S | Z = 0) , (~1 + S | Z = 1)")    
    theta <- try(glm.fit(S[Z==0,,drop = FALSE],Y[Z==0],family = binomial(link = link$name))$coefficients,silent = TRUE)
    if(!inherits(theta,'try-error'))Start$strat$theta[1:s] <- theta
    theta <- try(glm.fit(S[Z==1,,drop = FALSE],Y[Z==1],family = binomial(link = link$name))$coefficients,silent = TRUE)
    if(!inherits(theta,'try-error'))Start$strat$theta[-(1:s)] <- theta
    
    if(!any(is.na(Start$strat$theta))){
        loglik <- try(negloglik.tsm(Start$strat$theta,parsen = parsen,Y,Z,S,linkinv = linkinv,riskfun = risk.theta,verbose = verbose),silent = TRUE)
        if(!inherits(loglik,'try-error'))Start$strat$loglik <- -loglik 
    }

    loglik <- rep(-Inf,length(Start))    
    for (i in 1:length(Start)) if(is.numeric(Start[[i]]$loglik)) loglik[i] <- Start[[i]]$loglik
    if(all(!is.finite(loglik))) return(Start)
    c(list(best = Start[[which.max(loglik)]]$theta),Start)    
}

tsm.fit <- function(Y,Z,S,fpr,threshold,start = NULL,parsen = NULL,
constrained = TRUE,link = make.link('logit'),linkinv = NULL,
optimfun = optim,control = list(maxit = 1e5),intercept = TRUE,
est.scale = c("c","r"),NAMES = NULL,verbose = 0){
# this block of code is common with tsm.boot except for what it returns in case optim() failed
############################################################################################   
    fit <- NULL  
    Start <- NULL
    alpha <- NULL
    theta <- NULL
    if(!length(linkinv))linkinv <- link$linkinv     
    # 'unconstrained' glm() estimate on risk model scale  
    if(!constrained){
        # 'risk-scale'
        warning("With constraints = FALSE, the estimate Q = Risk(Z=0,S) - Risk(Z=1,S) may be less than zero for some of its entries thus estimates of classification accuracy such as the ROC curve and AUC are not reliable.")
        fit <- try(glm.fit(cbind(S,Z*S),Y,family = binomial(link = link$name),start = start,intercept = intercept),silent = TRUE)
        if (inherits(fit,'try-error')) return(fit)
        para <- coef(fit);attr(para,'NAMES') <- NAMES
        alpha <- alpha.theta(para) 
        theta <- theta.alpha(alpha) 
    }else{      
        # initial values ranked by likelihood, start is on 'risk' scale
        if(is.null(start)){
            Start <- tsm.start(Y,Z,S,link = link,linkinv = linkinv,verbose = verbose)
            start <- Start$best
            Start$best <- NULL
        }
        
        riskfun <- if(est.scale == "c") tsm::risk.alpha else tsm::risk.theta
        if(est.scale == "c") start <- alpha.theta(start) #convert to "cumulative link model" scale
        fit <- try(optimfun(c(start),negloglik.tsm,control = control,parsen = parsen,Y=Y,Z=Z,S=S,
                            linkinv = linkinv,riskfun = riskfun,verbose = verbose))
        if (inherits(fit,'try-error')){return(fit)}
        if (fit$convergence!=0){warning("optim failed to converge");return(fit)}
        para <- fit$par;attr(para,'NAMES') <- NAMES        
        if(est.scale == "c"){
            theta <- theta.alpha(para)
            alpha <- alpha.theta(theta)     
        }else{
            alpha <- alpha.theta(para)
            theta <- theta.alpha(alpha)             
        }
    }
    risk  <- risk.alpha(S,alpha,linkinv)
  	Q     <- c(risk[,1] - risk[,2]) #risk0 - risk1
    # if(any(Q < 0)){
        # txt <- "Negative entries found in risk difference Q = Risk(Z=0,S) - Risk(Z=1,S) which should be nonnegative. "
        # if(!length(parsen)){
            # txt<-c(txt,"Setting (phi,gamma) = (0,Inf,..,Inf).")
            # parsen <- c(theta)*0.0;parsen[-1] <- Inf
        # }
        # warning(txt)
    # }    
	measures <- get.measures(risk,S = S,parsen = parsen,fpr = fpr,threshold = threshold)
    ############################################################################################ 

# fit now has 'Q', roc(fpr),... see get.measures() 
    coefs <- c(theta)
    if(length(NAMES))names(coefs) <- c(NAMES[-1],NAMES[1],paste(NAMES[-(1:2)],":",NAMES[1],sep = ""))
    fit <- c(list(coefficients = coefs),measures,list(optim = fit))

# Keep 'matrix' form of parameters theta and alpha
    fit$theta <- theta #corresponds to [(1,S),  ,Z*(1,S)]
    fit$alpha <- alpha #corresponds to [(1,S) (1,S)] in two categories	
    
# bootstrapped variance    
# names of bootstrapped parameters - used in summary.tsm()        
	fit$boot.names <- c("coefficients","alpha")
    if(!length(parsen))fit$boot.names <- c(fit$boot.names,"auc","tpf.q","fpf.q","Rinvp.q","roc") # fit 
    else fit$boot.names <- c(fit$boot.names,"auc_u","auc_n","auc_h","roc_u","roc_n","roc_h") # sensitivity analysis
    fit$boot.length <- sum(unlist(lapply(fit[match(fit$boot.names,names(fit))],length))) 
    fit$cov    <- rep(NA,fit$boot.length) 
    
    fit$risk   <- risk;if(length(NAMES))colnames(fit$risk) <- paste(NAMES[1],0:1,sep="=")
    fit$fitted <- numeric(length(Z))
    fit$fitted[as.logical(Z)] <- fit$risk[as.logical(Z),2] 
    fit$fitted[!as.logical(Z)] <- fit$risk[!as.logical(Z),1] 
	fit$cdfq   <- ecdf(fit$Q)
    if(length(parsen)){
        fit$roc_u.fun <- try(approxfun(fit$FPFu.Q,fit$TPFn.Q),TRUE)
        fit$roc_n.fun <- try(approxfun(fit$FPFn.Q,fit$TPFn.Q),TRUE)
        fit$roc_h.fun <- try(approxfun(fit$FPFh.Q,fit$TPFn.Q),TRUE)
        fit$tpf_n.fun <- try(approxfun(fit$Q,fit$TPFn.Q),TRUE)
        fit$fpf_n.fun <- try(approxfun(fit$Q,fit$FPFn.Q),TRUE)
        fit$fpf_u.fun <- try(approxfun(fit$Q,fit$FPFu.Q),TRUE)    
    }else{
        fit$roc.fun <- try(approxfun(fit$FPF.Q,fit$TPF.Q),TRUE)
        fit$tpf.fun <- try(approxfun(fit$Q,fit$TPF.Q),TRUE)
        fit$fpf.fun <- try(approxfun(fit$Q,fit$FPF.Q),TRUE)
    }
    

    if(length(Start)){
        for(i in 1:length(Start))names(Start[[i]]$theta) <- names(fit$coefficients)
        fit$Start <- Start;
    }
    if(length(start)){fit$start <- start;names(fit$start) <- names(fit$coefficients)}
    fit$estimation.scale <- ifelse(est.scale == "c",
        "cumulative link model for joint distribution of potential outcomes",
        "risk model")
    fit$link <- link
    fit
}

# to be called by boot()
# output: unlist()-ed vector of (theta,alpha,auc,tpf(threshold),fpf(threshold),Rinvp(threshold),roc(fpr))
tsm.boot <- function(data,index,Y,Z,S,fpr,threshold,
start = NULL,parsen = NULL,constrained = TRUE,link = make.link('logit'),linkinv = NULL,
optimfun = optim,control = list(maxit = 1e5),boot.names = NULL,boot.length = 1,intercept = TRUE,
est.scale = c("c","r"),verbose = 0){
    Y <- Y[index];Z <- Z[index];S <- S[index,,drop = FALSE]
############################################################################################         
    
    fit <- NULL  
    Start <- list()
    alpha.start <- alpha <- NULL
    theta.start <- theta <- NULL
    if(!length(linkinv))linkinv <- link$linkinv 
    
    # 'unconstrained' glm() estimate on risk model scale  
    if(!constrained){
        # 'risk-scale'
        warning("With constraints = FALSE, the estimate of the probability of a potential outcome belonging to one of the categories may be less than zero, hence estimates of classification accuracy such as the ROC curve and AUC are not reliable.")
        fit <- try(glm.fit(cbind(S,Z*S),Y,family = binomial(link = link$name),start = start,intercept = intercept),silent = TRUE)
        if (inherits(fit,'try-error')) fit$coefficients <- numeric(2 * NCOL(S)) * NA
        alpha <- alpha.theta(coef(fit)) 
        theta <- theta.alpha(alpha) 
    }else{      
        # initial values ranked by likelihood
        if(is.null(start)){
            Start <- tsm.start(Y,Z,S,parsen = NULL,link = link,linkinv = linkinv,verbose = verbose)
            start <- Start$best
            Start$best <- NULL
            if(est.scale == "c") start <- alpha.theta(start) 
        }
        riskfun <- if(est.scale == "c") tsm::risk.alpha else tsm::risk.theta
        if(est.scale == "c") start <- alpha.theta(start) #convert to "cumulative link model" scale
        fit <- try(optimfun(c(start),negloglik.tsm,control = control,parsen = parsen,Y=Y,Z=Z,S=S,
                       linkinv = linkinv,riskfun = riskfun,verbose = verbose))
        if (inherits(fit,'try-error')){fit$par <- numeric(2 * NCOL(S)) * NA}
        if (fit$convergence!=0)fit$par <- numeric(2 * NCOL(S)) * NA
        if(est.scale == "c"){
            theta <- theta.alpha(fit$par)
            alpha <- alpha.theta(theta)     
        }else{
            alpha <- alpha.theta(fit$par)
            theta <- theta.alpha(alpha)             
        }        
    }
# this length must match length(res) below!
    if(any(is.na(alpha)))  return(rep(NA,boot.length))
    risk  <- risk.alpha(S,alpha,linkinv)
  	Q     <- c(risk[,1] - risk[,2])
	measures <- get.measures(risk,S = S,parsen = parsen,fpr = fpr,threshold = threshold)
    
############################################################################################ 
  
    if(length(boot.names)) measures <- measures[match(boot.names,names(measures),nomatch = 0)]    
    res <- c(theta,alpha,unlist(measures))
    return(res)
}

# formula = 'response ~ S1 + .. + Sd for markers (S1,..,Sd). 
# treatment = right-handed formula denoting treatment, ~Z if 'Z' indicates treatment assignment
tsm<-function (formula, data, treatment,subset, na.action,est.scale = c("risk","clpo"),
start = NULL,parsen = NULL, constrained = TRUE,link = c("logit","probit"),
fpr = seq(.1,.9,by=.1),q = c(.035,.085),signif.level = NULL,
nboot = 0,seed = NULL,
model = FALSE, x = TRUE, y = TRUE,
optimfun = c("optim","nmk","hjk"),control = list(),
parallel = c("no", "multicore", "snow"),
ncpus = getOption("boot.ncpus", 1L),cl = NULL
,verbose = 0,
...
)
{

## rounding at 15th digit because function delta.phi_gamma() gives different results if the linkinv
## function it uses is given as 'expit' defined in R vs '$linkinv" function in binomial logit-link family:
## |expit(Inf) - logit$linkinv(Inf)| ~= .Machine$double.eps = 2.220446e-16 on this machine. 
## This causes a problem in uniroot because for f = delta.phi_gamma(x) = function(x){mean(linkinv(x + S_gamma) * risk1)-phi_rho1} 
## f(-100) < f(100) < 0 for linkinv = logit$linkinv (so uniroot fails) but
## f(-100) < 0 < f(100) = 0 for linkinv = expit defined in R and thus uniroot$solution = 100
## see delta.phi_gamma(x) 
#  options(radix10.round.digits = as.integer(.Machine$double.digits * log(.Machine$double.base) / log(10.0)))

    
# as in lm()	
    if (missing(data)) data <- environment(formula)
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- as.name("model.frame")
    mf <- eval(mf, parent.frame())
    mt <- attr(mf, "terms")
    Y <- model.response(mf, "any")
    if (length(dim(Y)) == 1L) {
        nm <- rownames(Y)
        dim(Y) <- NULL
        if (!is.null(nm)) 
            names(Y) <- nm
    }
   
# intercept-free model is not supported but argument is passed through    
    intercept <- attr(mt, "intercept") > 0L
    if(!intercept) stop('No-intercept model is not supported.')

# link	
    link <- make.link(link[1])
    arg <- list(...)
    linkinv <- arg$linkinv
    if(is.character(linkinv))linkinv <- get(linkinv)    
    
# scale of estimation on parameter of the cumulative link model (as opposed to the risk model)
    est.scale <- substr(est.scale[1],1,1)
    
# treatment 1/0 or T/F
    Z <- treatment
    if(typeof(Z) == "language") Z <- model.matrix(update(Z,~-1+.),data)
    NAMES <- colnames(Z)
    if(!length(NAMES)) NAMES <- names(Z)
    if(!length(NAMES)) NAMES <- c('tx')
    Z <- as.numeric(c(Z))
    
# marker	
    S <- if (!is.empty.model(mt)) model.matrix(mt, mf, contrasts) else matrix(, NROW(Y), 0L)

# NAMES = (name.of.treatment,name.of.marker.1,..,name.of.marker.k)    
    NAMES <- c(NAMES,colnames(S))

# select optimizer    
    optimfun <- try(get(optimfun[1]),TRUE)
    if(inherits(optimfun,'try-error')) optimfun <- optim
        

# construct a formula that contains interaction Z * S if constrained = FALSE (e.g. glm())    
#    fm <- paste(".~","(.)*",as.character(treatment)[2],sep = "")
#    fm <- do.call(update.formula,list(formula,fm))
#	 formula.interaction <- fm
    
# Estimation: return a list of length 0 or a list of numeric vectors. 
    fit <- tsm.fit(Y,Z,S,fpr,q,start = start,parsen = parsen,constrained = constrained,
    link = link,linkinv = linkinv,control = control,intercept = intercept,
    est.scale = est.scale,NAMES = NAMES,verbose = verbose)
    if(inherits(fit,'try-error')) return(fit)   
    

# Stratified (Z) bootstrap
    if(nboot){
#        if(nboot < 2)nboot <- 2
        if(length(seed)) set.seed(seed[1])
        fit$boot <- try(boot(data = 1:NROW(Y),statistic = tsm.boot,R = nboot,strata = Z,
                     Y = Y,Z = Z,S = S,start = start,parsen = parsen,constrained = constrained,
                     fpr = fpr,threshold = q,
                     link = link,linkinv = linkinv,
                     control = control,
                     boot.names = fit$boot.names,boot.length = fit$boot.length,
                     intercept = intercept, #for glm fit
                     parallel = parallel[1], ncpus = ncpus,cl = cl,
                     est.scale = est.scale,verbose = verbose),TRUE)
                     
        if(!inherits(fit$boot,'try-error')){
            fit$cov  <- rowMeans((fit$boot$t0 - t(fit$boot$t))^2,na.rm = TRUE) 
            if(length(signif.level))fit$ci   <- t(apply(fit$boot$t,2,FUN = quantile,probs = c(0.5*signif.level,rev(1-0.5*signif.level)),type = 1,na.rm = TRUE))
            }
    }
    fit$parsen <- parsen
    fit$nboot<- nboot
    fit <- c(fit,list(fpr = fpr,q = q,link = link,
                      linkinv = if(is.null(linkinv))link$linkinv else linkinv,
                      signif.level = signif.level
                      )
            )
            
    fit <- c(fit,list(call = match.call(), formula = formula,treatment = treatment,na.action = attr(mf, "na.action")))
    if (model) fit$model <- mf
    if (y){ fit <- c(fit,list(y=Y))}
    if (x){ fit <- c(fit,list(x=S,z=Z))}        
#   if(details)fit<-c(fit,list(terms = mt,xlevels = .getXlevels(mt, mf),contrasts = attr(x, "contrasts")))
    class(fit) <- c(fit$class, c("tsm"))
    fit
}

summary.tsm <- 
function (fit,signif.level = NULL,...) 
{
    z <- fit
    if (!inherits(z, "tsm")) stop("invalid 'tsm' object")
    ans <- list()
    ans$aliased <- is.na(coef(z))
    if (!is.null(z$na.action)) ans$na.action <- z$na.action
    ans <- c(z[c("call")],ans)

    if(length(signif.level) & (z$nboot > 0)){     
        ci <- t(apply(z$boot$t,2,FUN = quantile,probs = c(0.5*signif.level,rev(1.0-0.5*signif.level)),type = 1,na.rm = TRUE))
    }else{
        ci <- fit$ci
    }
    
    # extract bootstrapped 'se' and 'quantiles' of unlist(tsm.fit())
    index <- match(z$boot.names,names(z)) 

    from <- 1;to <- 1
    for(i in 1:length(index)){
       est <- c(z[[index[i]]]);len <- length(est)
       to <- from + len	- 1
       se <- sqrt(z$cov[from:to])
       ans[[i]] <- cbind(est, se)
       dimnames(ans[[i]]) <- list(names(est),c("Estimate", "Std. Error"))
       if(length(ci))ans[[i]] <- cbind(ans[[i]],ci[from:to,,drop = FALSE])		
       names(ans)[i] <- names(z)[index[i]]
       from <- from + len	   
    }
    
# set summary rownames at input risk difference threshold 'q'    
    thr <- paste(round(z$q,digits = 4))
    index <- match(c("tpf.q","fpf.q","Rinvp.q"),names(ans),nomatch = 0)
    for(i in index)if(i)rownames(ans[[i]]) <- thr

# set roc rownames at input fpr         
	rn <- paste(round(z$fpr,digits = 4))
    index <- match(c("roc","roc_u","roc_n","roc_h"),names(ans),nomatch = 0)
    for(i in index)if(i)rownames(ans[[i]]) <- rn

    class(ans) <- "summary.tsm"
    ans
}

plot.roc <- function(fit,what = c("roc","n","u","h"),...){
            arg <- list(...)
            main <- arg$main
            
            if(is.null(main)) main <- "ROC curve"

            what <- unlist(lapply(what,substr,1,1))

            index <- pmatch(what,c("r","n","u","h"),nomatch = 0)
            index <- index[index != 0]
            w  <- c("roc","roc_n","roc_u","roc_h")[index]
            wname <- c("roc","unaffected or harmed vs benefitted","unaffected vs benefitted","harmed vs benefitted")[index]
            sfit <- summary.tsm(fit,...)
            index <- pmatch(w,names(sfit),nomatch = 0)

            wname <- wname[index != 0]
            index <- index[index != 0]
            if(!sum(index)) stop("Invalid 'what' argument for plot.roc")
            
            plot(c(0,1),c(0,1), type="n", xlim=c(0,1), ylim=c(0,1), xlab="FPR", ylab="TPR", cex.lab=1.5, 
            frame.plot=TRUE, axes=FALSE,main = main)
            axis(1, at=c(0,1))
            axis(2, at=c(0,1))
            abline(h=seq(0,1,0.2), v=seq(0,1,0.2), col="lightgray", lty="solid", lwd=par("lwd"))
            abline(0, 1, col=8)
           

            #[estimate , a1,a2,..,ap , 1-ap,..,1-a2,1-a1]
            nci <- length(arg$signif.level);if(!nci)nci<-length(fit$signif.level)
            nci <- nci * (fit$nboot > 1)
            lty.ci <- NULL
            if(nci){lty.ci <- 1:nci;lty.ci <- c(lty.ci,rev(lty.ci))}
            lty0 <- c(0,lty.ci) + 1
            
            x <- c(0,fit$fpr,1)
            y <- NULL;lty <- NULL 
            for(i in 1:length(index)){
                yi <- sfit[[index[i]]]
                yi <- yi[,-2,drop = FALSE]
                colnames(yi)[1] <- wname[i]
                y  <- cbind(y,yi)
                lty.i <- lty0
                if(i > 1) lty.i <- lty.i + lty[length(lty)]
                lty <- c(lty,lty.i)
            }
            y <- rbind(0,y,1)
            ny<- ncol(y) #[estimate,a1,a2,..,ap,1-ap,..,1-a2,1-a1, ..., ]           
            
            # lty  <- (1:ncol(y)) 
            # lty1 <- lty[-1];
            # lty1 <- lty1[1:floor(length(lty1)/2)]   
            # if(length(lty) > 2) lty <- c(lty[1],lty1,rev(lty1))

            lty <- if(length(arg$lty)) arg$lty else lty
            col <- if(length(arg$col))arg$col else rep("black",ny)
            lwd <- if(length(arg$lwd))arg$lwd else rep(1,ny)
            leg <- arg$legend;if(!length(leg)) leg <- colnames(y)
            for(i in 1:ny){if(!any(is.na(y[,i])))lines(x,y[,i],lty = lty[i],col = col[i],lwd = lwd[i])}
            if(length(leg))legend("bottomright",leg,lty = lty,col = col)		
}

plot.Q <- function(fit,what = c("cdf","tpf","fpf"),...){            
			arg <- list(...)
            main <- arg$main;
            
            x <- fit$Q[-1] #[-1]: ignoring the '-Inf' that prepended sort(unique(Q)) in get.measures() or get.measures.sen()
            rng <- range(x)
			plot(rng,c(0,1), type="n", xlim=rng, ylim=c(0,1), xlab="risk0(s) - risk1(s)", ylab="",
				 cex.lab=1.5, frame.plot=TRUE, axes=TRUE,main = main)

			y <- NULL
            what <- unlist(lapply(what,substr,1,1))
			if(any(what == "c")) y <- cbind(y,CDF_Q = fit$cdfq(x))
            if(any(what == "t")) y <- cbind(y,TPF = fit$TPF.Q[-1])
            if(any(what == "f")) y <- cbind(y,FPF = fit$FPF.Q[-1])

            ny<- ncol(y) 
            lty <- if(length(arg$lty)) arg$lty else 1:ny			
            col <- if(length(arg$col))arg$col else rep("black",ny)
            lwd <- if(length(arg$lwd))arg$lwd else rep(1,ny)
            leg <- arg$legend;if(!length(leg)) leg <- colnames(y)
            for(i in 1:ny)if(!any(is.na(y[,i])))lines(x,y[,i],lty = lty[i],col = col[i],lwd = lwd[i])
            if(length(leg))legend("right",leg,lty = lty,col = col)		
}
		
plot.tsm <- function(fit,what = c("roc","cdf","tpf","fpf","n","h","u"),...){
    if (!inherits(fit, "tsm")) stop("invalid 'tsm' object")
    w <- unlist(lapply(what,substr,1,1))
    what.roc <- what[pmatch(c('r','n','u','h'),w,nomatch = 0)] 
    what.Q   <- what[pmatch(c('c','t','f'),w,nomatch = 0)] 
        
    if(length(what.roc)){
        try.warn(plot.roc(fit,what.roc,...)) 
    }else if(length(what.Q))try.warn(plot.Q(fit,what.Q,...)) 
    
}

